# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""DolphinScheduler local resource plugin."""

import os
from pathlib import Path

from pydolphinscheduler.core.resource_plugin import ResourcePlugin
from pydolphinscheduler.exceptions import PyResPluginException


class Local(ResourcePlugin):
    """Local object, declare local resource plugin for task and workflow to dolphinscheduler.

    :param prefix: A string representing the prefix of Local.
    """

    # [start init_method]
    def __init__(self, prefix: str, *args, **kwargs):
        super().__init__(prefix, *args, **kwargs)

    # [end init_method]

    # [start read_file_method]
    def read_file(self, suf: str):
        """Get the content of the file.

        The address of the file is the prefix of the resource plugin plus the parameter suf.
        """
        path = Path(self.prefix).joinpath(suf)
        if not path.exists():
            raise PyResPluginException("{} is not found".format(str(path)))
        if not os.access(str(path), os.R_OK):
            raise PyResPluginException(
                "You don't have permission to access {}".format(self.prefix + suf)
            )
        with open(path, "r") as f:
            content = f.read()
        return content

    # [end read_file_method]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Task procedure."""

from typing import Dict

from pydolphinscheduler.constants import TaskType
from pydolphinscheduler.core.database import Database
from pydolphinscheduler.core.task import Task


class Procedure(Task):
    """Task Procedure object, declare behavior for Procedure task to dolphinscheduler.

    It should run database procedure job in multiply sql lik engine, such as:
    - ClickHouse
    - DB2
    - HIVE
    - MySQL
    - Oracle
    - Postgresql
    - Presto
    - SQLServer
    You provider datasource_name contain connection information, it decisions which
    database type and database instance would run this sql.
    """

    _task_custom_attr = {"method"}

    def __init__(self, name: str, datasource_name: str, method: str, *args, **kwargs):
        super().__init__(name, TaskType.PROCEDURE, *args, **kwargs)
        self.datasource_name = datasource_name
        self.method = method

    @property
    def task_params(self, camel_attr: bool = True, custom_attr: set = None) -> Dict:
        """Override Task.task_params for produce task.

        produce task have some specials attribute for task_params, and is odd if we
        directly set as python property, so we Override Task.task_params here.
        """
        params = super().task_params
        datasource = Database(self.datasource_name, "type", "datasource")
        params.update(datasource)
        return params



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Parse YAML file to create workflow."""

import logging
import os
import re
from pathlib import Path
from typing import Any, Dict

from pydolphinscheduler import configuration, tasks
from pydolphinscheduler.constants import Symbol
from pydolphinscheduler.core.parameter import ParameterType
from pydolphinscheduler.core.task import Task
from pydolphinscheduler.core.workflow import Workflow
from pydolphinscheduler.exceptions import PyDSTaskNoFoundException
from pydolphinscheduler.utils.yaml_parser import YamlParser

logger = logging.getLogger(__file__)

KEY_WORKFLOW = "workflow"
KEY_TASK = "tasks"
KEY_TASK_TYPE = "task_type"
KEY_DEPS = "deps"
KEY_OP = "op"

TASK_SPECIAL_KEYS = [KEY_TASK_TYPE, KEY_DEPS]


class ParseTool:
    """Enhanced parsing tools."""

    @staticmethod
    def parse_string_param_if_file(string_param: str, **kwargs):
        """Use $FILE{"data_path"} to load file from "data_path"."""
        if string_param.startswith("$FILE"):
            path = re.findall(r"\$FILE\{\"(.*?)\"\}", string_param)[0]
            base_folder = kwargs.get("base_folder", ".")
            path = ParseTool.get_possible_path(path, base_folder)
            with open(path, "r") as read_file:
                string_param = "".join(read_file)
        return string_param

    @staticmethod
    def parse_string_param_if_env(string_param: str, **kwargs):
        """Use $ENV{env_name} to load environment variable "env_name"."""
        if "$ENV" in string_param:
            key = re.findall(r"\$ENV\{(.*?)\}", string_param)[0]
            env_value = os.environ.get(key, "$%s" % key)
            string_param = string_param.replace("$ENV{%s}" % key, env_value)
        return string_param

    @staticmethod
    def parse_string_param_if_config(string_param: str, **kwargs):
        """Use ${CONFIG.var_name} to load variable "var_name" from configuration."""
        if "${CONFIG" in string_param:
            key = re.findall(r"\$\{CONFIG\.(.*?)\}", string_param)[0]
            if hasattr(configuration, key):
                string_param = getattr(configuration, key)
            else:
                string_param = configuration.get_single_config(key)

        return string_param

    @staticmethod
    def parse_string_param_if_parameter(string_param: str, **kwargs):
        """Use TYPE(value) to set local params."""
        key_path = kwargs.get("key_path")
        if key_path.split(Symbol.POINT)[0] not in {"input_params", "output_params"}:
            return string_param

        if not isinstance(string_param, str):
            return string_param

        result = re.findall(r"^(.*?)\((.*?)\)", string_param)
        if len(result) == 1 and len(result[0]) == 2:
            type_ = result[0][0].rstrip()
            value = result[0][1].rstrip()
            return ParameterType.type_sets[type_](value)
        else:
            return string_param

    @staticmethod
    def get_possible_path(file_path, base_folder):
        """Get file possible path.

        Return new path if file_path is not exists, but base_folder + file_path exists
        """
        possible_path = file_path
        if not Path(file_path).exists():
            new_path = Path(base_folder).joinpath(file_path)
            if new_path.exists():
                possible_path = new_path
                logger.info(f"{file_path} not exists, convert to {possible_path}")

        return possible_path


def get_task_cls(task_type) -> Task:
    """Get the task class object by task_type (case compatible)."""
    # only get task class from tasks.__all__
    all_task_types = {type_.capitalize(): type_ for type_ in tasks.__all__}
    task_type_cap = task_type.capitalize()
    if task_type_cap not in all_task_types:
        raise PyDSTaskNoFoundException("cant not find task %s" % task_type)

    standard_name = all_task_types[task_type_cap]
    return getattr(tasks, standard_name)


class YamlWorkflow(YamlParser):
    """Yaml parser for create workflow.

    :param yaml_file: yaml file path.

        examples1 ::

            parser = YamlParser(yaml_file=...)
            parser.create_workflow()

        examples2 ::

            YamlParser(yaml_file=...).create_workflow()

    """

    _parse_rules = [
        ParseTool.parse_string_param_if_file,
        ParseTool.parse_string_param_if_env,
        ParseTool.parse_string_param_if_config,
        ParseTool.parse_string_param_if_parameter,
    ]

    def __init__(self, yaml_file: str):
        with open(yaml_file, "r") as f:
            content = f.read()

        self._base_folder = Path(yaml_file).parent
        content = self.prepare_refer_workflow(content)
        super().__init__(content)

    def create_workflow(self):
        """Create workflow main function."""
        # get workflow parameters with key "workflow"
        workflow_params = self[KEY_WORKFLOW]

        # pop "run" parameter, used at the end
        is_run = workflow_params.pop("run", False)

        # use YamlWorkflow._parse_rules to parse special value of yaml file
        workflow_params = self.parse_params(workflow_params)

        workflow_name = workflow_params["name"]
        logger.info(f"Create workflow: {workflow_name}")
        with Workflow(**workflow_params) as workflow:
            # save dependencies between tasks
            dependencies = {}

            # save name and task mapping
            name2task = {}

            # get task datas with key "tasks"
            for task_data in self[KEY_TASK]:
                task = self.parse_task(task_data, name2task)

                deps = task_data.get(KEY_DEPS, [])
                if deps:
                    dependencies[task.name] = deps
                name2task[task.name] = task

            # build dependencies between task
            for downstream_task_name, deps in dependencies.items():
                downstream_task = name2task[downstream_task_name]
                for upstream_task_name in deps:
                    upstream_task = name2task[upstream_task_name]
                    upstream_task >> downstream_task

            workflow.submit()
            # if set is_run, run the workflow after submit
            if is_run:
                logger.info(f"run workflow: {workflow}")
                workflow.run()

        return workflow_name

    def parse_params(self, params: Any, key_path=""):
        """Recursively resolves the parameter values.

        The function operates params only when it encounters a string; other types continue recursively.
        """
        if isinstance(params, str):
            for parse_rule in self._parse_rules:
                params_ = params
                params = parse_rule(
                    params, base_folder=self._base_folder, key_path=key_path
                )
                if params_ != params:
                    logger.info(f"parse {params_} -> {params}")

        elif isinstance(params, list):
            for index in range(len(params)):
                params[index] = self.parse_params(params[index], key_path)

        elif isinstance(params, dict):
            for key, value in params.items():
                if not key_path:
                    new_key_path = key
                else:
                    new_key_path = key_path + Symbol.POINT + key
                params[key] = self.parse_params(value, new_key_path)

        return params

    @classmethod
    def parse(cls, yaml_file: str):
        """Recursively resolves the parameter values.

        The function operates params only when it encounters a string; other types continue recursively.
        """
        workflow_name = cls(yaml_file).create_workflow()
        return workflow_name

    def prepare_refer_workflow(self, content):
        """Allow YAML files to reference workflow derived from other YAML files."""
        workflow_paths = re.findall(r"\$WORKFLOW\{\"(.*?)\"\}", content)
        for workflow_path in workflow_paths:
            logger.info(
                f"find special token {workflow_path}, load workflow form {workflow_path}"
            )
            possible_path = ParseTool.get_possible_path(
                workflow_path, self._base_folder
            )
            workflow_name = YamlWorkflow.parse(possible_path)
            content = content.replace('$WORKFLOW{"%s"}' % workflow_path, workflow_name)

        return content

    def parse_task(self, task_data: dict, name2task: Dict[str, Task]):
        """Parse various types of tasks.

        :param task_data: dict.
                {
                    "task_type": "Shell",
                    "params": {"name": "shell_task", "command":"ehco hellp"}
                }

        :param name2task: Dict[str, Task]), mapping of task_name and task


        Some task type have special parse func:
            if task type is Switch, use parse_switch;
            if task type is Condition, use parse_condition;
            if task type is Dependent, use parse_dependent;
            other, we pass all task_params as input to task class, like "task_cls(**task_params)".
        """
        task_type = task_data["task_type"]
        # get params without special key
        task_params = {k: v for k, v in task_data.items() if k not in TASK_SPECIAL_KEYS}

        task_cls = get_task_cls(task_type)

        # use YamlWorkflow._parse_rules to parse special value of yaml file
        task_params = self.parse_params(task_params)

        if task_cls == tasks.Switch:
            task = self.parse_switch(task_params, name2task)

        elif task_cls == tasks.Condition:
            task = self.parse_condition(task_params, name2task)

        elif task_cls == tasks.Dependent:
            task = self.parse_dependent(task_params, name2task)

        else:
            task = task_cls(**task_params)
        logger.info(task_type, task)
        return task

    def parse_switch(self, task_params, name2task):
        """Parse Switch Task.

        This is an example Yaml fragment of task_params

        name: switch
        condition:
          - ["${var} > 1", switch_child_1]
          - switch_child_2
        """
        from pydolphinscheduler.tasks.switch import (
            Branch,
            Default,
            Switch,
            SwitchCondition,
        )

        condition_datas = task_params["condition"]
        conditions = []
        for condition_data in condition_datas:
            assert "task" in condition_data, "task must be in %s" % condition_data
            task_name = condition_data["task"]
            condition_string = condition_data.get("condition", None)

            # if condition_string is None, for example: {"task": "switch_child_2"}, set it to Default branch
            if condition_string is None:
                conditions.append(Default(task=name2task.get(task_name)))

            # if condition_string is not None, for example:
            # {"task": "switch_child_2", "condition": "${var} > 1"} set it to Branch
            else:
                conditions.append(
                    Branch(condition_string, task=name2task.get(task_name))
                )

        switch = Switch(
            name=task_params["name"], condition=SwitchCondition(*conditions)
        )
        return switch

    def parse_condition(self, task_params, name2task):
        """Parse Condition Task.

        This is an example Yaml fragment of task_params

        name: condition
        success_task: success_branch
        failed_task: fail_branch
        OP: AND
        groups:
          -
            OP: AND
            groups:
              - [pre_task_1, true]
              - [pre_task_2, true]
              - [pre_task_3, false]
          -
            OP: AND
            groups:
              - [pre_task_1, false]
              - [pre_task_2, true]
              - [pre_task_3, true]

        """
        from pydolphinscheduler.tasks.condition import (
            FAILURE,
            SUCCESS,
            And,
            Condition,
            Or,
        )

        def get_op_cls(op):
            cls = None
            if op.lower() == "and":
                cls = And
            elif op.lower() == "or":
                cls = Or
            else:
                raise Exception("OP must be in And or Or, but get: %s" % op)
            return cls

        second_cond_ops = []
        for first_group in task_params["groups"]:
            second_op = first_group["op"]
            task_ops = []
            for condition_data in first_group["groups"]:
                assert "task" in condition_data, "task must be in %s" % condition_data
                assert "flag" in condition_data, "flag must be in %s" % condition_data
                task_name = condition_data["task"]
                flag = condition_data["flag"]
                task = name2task[task_name]

                # for example: task = pre_task_1, flag = true
                if flag:
                    task_ops.append(SUCCESS(task))
                else:
                    task_ops.append(FAILURE(task))

            second_cond_ops.append(get_op_cls(second_op)(*task_ops))

        first_op = task_params["op"]
        cond_operator = get_op_cls(first_op)(*second_cond_ops)

        condition = Condition(
            name=task_params["name"],
            condition=cond_operator,
            success_task=name2task[task_params["success_task"]],
            failed_task=name2task[task_params["failed_task"]],
        )
        return condition

    def parse_dependent(self, task_params, name2task):
        """Parse Dependent Task.

        This is an example Yaml fragment of task_params

        name: dependent
        denpendence:
        OP: AND
        groups:
          -
            OP: Or
            groups:
              - [pydolphin, task_dependent_external, task_1]
              - [pydolphin, task_dependent_external, task_2]
          -
            OP: And
            groups:
              - [pydolphin, task_dependent_external, task_1, LAST_WEDNESDAY]
              - [pydolphin, task_dependent_external, task_2, last24Hours]

        """
        from pydolphinscheduler.tasks.dependent import (
            And,
            Dependent,
            DependentDate,
            DependentItem,
            Or,
        )

        def workflow_dependent_date(dependent_date):
            """Parse dependent date (Compatible with key and value of DependentDate)."""
            dependent_date_upper = dependent_date.upper()
            if hasattr(DependentDate, dependent_date_upper):
                dependent_date = getattr(DependentDate, dependent_date_upper)
            return dependent_date

        def get_op_cls(op):
            cls = None
            if op.lower() == "and":
                cls = And
            elif op.lower() == "or":
                cls = Or
            else:
                raise Exception("OP must be in And or Or, but get: %s" % op)
            return cls

        def create_dependent_item(source_items):
            """Parse dependent item.

            project_name: pydolphin
            workflow_name: task_dependent_external
            dependent_task_name: task_1
            dependent_date: LAST_WEDNESDAY
            """
            project_name = source_items["project_name"]
            workflow_name = source_items["workflow_name"]
            dependent_task_name = source_items["dependent_task_name"]
            dependent_date = source_items.get("dependent_date", DependentDate.TODAY)
            dependent_item = DependentItem(
                project_name=project_name,
                workflow_name=workflow_name,
                dependent_task_name=dependent_task_name,
                dependent_date=workflow_dependent_date(dependent_date),
            )

            return dependent_item

        second_dependences = []
        for first_group in task_params["groups"]:
            second_op = first_group[KEY_OP]
            dependence_items = []
            for source_items in first_group["groups"]:
                dependence_items.append(create_dependent_item(source_items))

            second_dependences.append(get_op_cls(second_op)(*dependence_items))

        first_op = task_params[KEY_OP]
        dependence = get_op_cls(first_op)(*second_dependences)

        task = Dependent(
            name=task_params["name"],
            dependence=dependence,
        )
        return task


def create_workflow(yaml_file):
    """CLI."""
    YamlWorkflow.parse(yaml_file)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Task Python."""

import logging
import re
import types
from pathlib import Path
from typing import Union

from stmdency.extractor import Extractor

from pydolphinscheduler.constants import TaskType
from pydolphinscheduler.core.task import Task
from pydolphinscheduler.exceptions import PyDSParamException

log = logging.getLogger(__file__)


class Python(Task):
    """Task Python object, declare behavior for Python task to dolphinscheduler.

    Python task support two types of parameters for :param:``definition``, and here is an example:

    Using str type of :param:``definition``

    .. code-block:: python

        python_task = Python(name="str_type", definition="print('Hello Python task.')")

    Or using Python callable type of :param:``definition``

    .. code-block:: python

        def foo():
            print("Hello Python task.")

        python_task = Python(name="str_type", definition=foo)

    :param name: The name for Python task. It define the task name.
    :param definition: String format of Python script you want to execute or Python callable you
        want to execute.
    """

    _task_custom_attr = {"raw_script"}

    ext: set = {".py"}
    ext_attr: Union[str, types.FunctionType] = "_definition"

    def __init__(
        self, name: str, definition: Union[str, types.FunctionType], *args, **kwargs
    ):
        self._definition = definition
        super().__init__(name, TaskType.PYTHON, *args, **kwargs)

    def _build_exe_str(self) -> str:
        """Build executable string from given definition.

        Attribute ``self.definition`` almost is a function, we need to call this function after parsing it
        to string. The easier way to call a function is using syntax ``func()`` and we use it to call it too.
        """
        definition = getattr(self, "definition")
        if isinstance(definition, types.FunctionType):
            loc = definition.__code__.co_filename
            extractor = Extractor(Path(loc).open("r").read())
            stm = extractor.get_code(definition.__name__)
            func_str = f"{stm}{definition.__name__}()"
        else:
            pattern = re.compile("^def (\\w+)\\(")
            find = pattern.findall(definition)
            if not find:
                return definition
            # Keep function str and function callable always have one blank line
            func_str = (
                f"{definition}{find[0]}()"
                if definition.endswith("\n")
                else f"{definition}\n{find[0]}()"
            )
        return func_str

    @property
    def raw_script(self) -> str:
        """Get python task define attribute `raw_script`."""
        if isinstance(getattr(self, "definition"), (str, types.FunctionType)):
            return self._build_exe_str()
        else:
            raise PyDSParamException(
                "Parameter definition do not support % for now.",
                type(getattr(self, "definition")),
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# [start workflow_declare]
r"""
A tutorial example set local parameter in pydolphinscheduler.

Method 1:
    task = Shell(..., input_params={"input":"a"}, output_params={"output": "b"})

Method 2:
    task = Shell(...)
    task.add_in("input", "a")
    task.add_out("output", "b")
"""

from pydolphinscheduler.core.parameter import ParameterType
from pydolphinscheduler.core.workflow import Workflow
from pydolphinscheduler.tasks.shell import Shell

with Workflow(name="local_parameter_example", release_state="offline") as workflow:
    # [start parameter example]
    # define a parameter "a", and use it in Shell task
    example1_input_params = Shell(
        name="example1_input_params",
        command="echo ${a}",
        input_params={
            "a": "123",
        },
    )

    # define a parameter "random_value", and pass it to the downstream tasks
    example2_output_params = Shell(
        name="example2_output_params",
        command="""
        val=$(echo $RANDOM)
        echo "#{setValue(random_value=${val})}"
        echo $val
        """,
        output_params={
            "random_value": "",
        },
    )

    # use the parameter "random_value", from upstream tasks
    # we don't need to define input_params again if the parameter is from upstram tasks
    example2_input_params = Shell(
        name="example2_input_params", command="""echo ${random_value}"""
    )

    example2_output_params >> example2_input_params
    # [end parameter example]

    # [start parameter define]
    # Add parameter via task arguments
    task_1 = Shell(
        name="task_1",
        command="echo hello pydolphinscheduler",
        input_params={
            "value_VARCHAR": "abc",
            "value_INTEGER": 123,
            "value_FLOAT": 0.1,
            "value_BOOLEAN": True,
        },
        output_params={
            "value_EMPTY": None,
        },
    )

    # Add parameter via task instance's method
    task_2 = Shell(name="task_2", command="echo hello pydolphinscheduler")

    task_2.add_in("value_VARCHAR", "abc")
    task_2.add_in("value_INTEGER", 123)
    task_2.add_in("value_FLOAT", 0.1)
    task_2.add_in("value_BOOLEAN", True)
    task_2.add_out("value_EMPTY")

    # Task 1 is the same as task 2

    # Others parameter types which cannot be converted automatically, must declare type explicitly
    task_3 = Shell(
        name="task_3",
        command="echo '123' >> test.txt",
        input_params={
            "value_LONG": ParameterType.LONG("1000000"),
            "value_DATE": ParameterType.DATE("2022-01-02"),
            "value_TIME": ParameterType.TIME("2022-01-01"),
            "value_TIMESTAMP": ParameterType.TIMESTAMP(123123124125),
            "value_LIST": ParameterType.LIST("123123"),
        },
        output_params={
            "output_INTEGER": ParameterType.INTEGER(100),
            "output_LIST": ParameterType.LIST(),
            "output_FILE": ParameterType.FILE("test.txt"),
        },
    )

    workflow.submit()
    # [end parameter define]
# [end workflow_declare]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Commands line interface's command of pydolphinscheduler."""

import click
from click import echo

import pydolphinscheduler
from pydolphinscheduler.configuration import (
    get_single_config,
    init_config_file,
    set_single_config,
)
from pydolphinscheduler.core.yaml_workflow import create_workflow

version_option_val = ["major", "minor", "micro"]


@click.group()
def cli():
    """Apache DolphinScheduler Python API's command line interface."""


@cli.command()
@click.option(
    "--part",
    "-p",
    required=False,
    type=click.Choice(version_option_val, case_sensitive=False),
    multiple=False,
    help="The part of version your want to get.",
)
def version(part: str) -> None:
    """Show current version of pydolphinscheduler."""
    if part:
        idx = version_option_val.index(part)
        echo(f"{pydolphinscheduler.__version__.split('.')[idx]}")
    else:
        echo(f"{pydolphinscheduler.__version__}")


@cli.command()
@click.option(
    "--init",
    "-i",
    is_flag=True,
    help="Initialize and create configuration file to `PYDS_HOME`.",
)
@click.option(
    "--set",
    "-s",
    "setter",
    multiple=True,
    type=click.Tuple([str, str]),
    help="Set specific setting to config file."
    "Use multiple ``--set <KEY> <VAL>`` options to set multiple configs",
)
@click.option(
    "--get",
    "-g",
    "getter",
    multiple=True,
    type=str,
    help="Get specific setting from config file."
    "Use multiple ``--get <KEY>`` options to get multiple configs",
)
def config(getter, setter, init) -> None:
    """Manage the configuration for pydolphinscheduler."""
    if init:
        init_config_file()
    elif getter:
        click.echo("The configuration query as below:\n")
        configs_kv = [f"{key} = {get_single_config(key)}" for key in getter]
        click.echo("\n".join(configs_kv))
    elif setter:
        for key, val in setter:
            set_single_config(key, val)
        click.echo("Set configuration done.")


@cli.command()
@click.option(
    "--yaml_file",
    "-f",
    required=True,
    help="YAML file path",
    type=click.Path(exists=True),
)
def yaml(yaml_file) -> None:
    """Create workflow using YAML file."""
    create_workflow(yaml_file)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Task Pytorch."""
from typing import Optional

from pydolphinscheduler.constants import TaskType
from pydolphinscheduler.core.task import Task


class DEFAULT:
    """Default values for Pytorch."""

    is_create_environment = False
    project_path = "."
    python_command = "${PYTHON_HOME}"


class Pytorch(Task):
    """Task Pytorch object, declare behavior for Pytorch task to dolphinscheduler.

    See also: `DolphinScheduler Pytorch Task Plugin
    <https://dolphinscheduler.apache.org/en-us/docs/dev/user_doc/guide/task/pytorch.html>`_

    :param name: task name
    :param script: Entry to the Python script file that you want to run.
    :param script_params: Input parameters at run time.
    :param project_path: The path to the project. Default "." .
    :param is_create_environment: is create environment. Default False.
    :param python_command: The path to the python command. Default "${PYTHON_HOME}".
    :param python_env_tool: The python environment tool. Default "conda".
    :param requirements: The path to the requirements.txt file. Default "requirements.txt".
    :param conda_python_version: The python version of conda environment. Default "3.7".
    """

    _task_custom_attr = {
        "script",
        "script_params",
        "other_params",
        "python_path",
        "is_create_environment",
        "python_command",
        "python_env_tool",
        "requirements",
        "conda_python_version",
    }

    def __init__(
        self,
        name: str,
        script: str,
        script_params: str = "",
        project_path: Optional[str] = DEFAULT.project_path,
        is_create_environment: Optional[bool] = DEFAULT.is_create_environment,
        python_command: Optional[str] = DEFAULT.python_command,
        python_env_tool: Optional[str] = "conda",
        requirements: Optional[str] = "requirements.txt",
        conda_python_version: Optional[str] = "3.7",
        *args,
        **kwargs,
    ):
        """Init Pytorch task."""
        super().__init__(name, TaskType.PYTORCH, *args, **kwargs)
        self.script = script
        self.script_params = script_params
        self.is_create_environment = is_create_environment
        self.python_path = project_path
        self.python_command = python_command
        self.python_env_tool = python_env_tool
        self.requirements = requirements
        self.conda_python_version = conda_python_version

    @property
    def other_params(self):
        """Return other params."""
        conds = [
            self.is_create_environment != DEFAULT.is_create_environment,
            self.python_path != DEFAULT.project_path,
            self.python_command != DEFAULT.python_command,
        ]
        return any(conds)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# [start workflow_declare]
"""A example workflow for task flink."""

from pydolphinscheduler.core.workflow import Workflow
from pydolphinscheduler.tasks.flink import DeployMode, Flink, ProgramType

with Workflow(name="task_flink_example", tenant="tenant_exists") as workflow:
    task = Flink(
        name="task_flink",
        main_class="org.apache.flink.streaming.examples.wordcount.WordCount",
        main_package="WordCount.jar",
        program_type=ProgramType.JAVA,
        deploy_mode=DeployMode.LOCAL,
    )
    workflow.run()
# [end workflow_declare]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""DolphinScheduler User object."""

from typing import Optional

from pydolphinscheduler import configuration
from pydolphinscheduler.java_gateway import gateway
from pydolphinscheduler.models import BaseSide, Tenant


class User(BaseSide):
    """DolphinScheduler User object."""

    _KEY_ATTR = {
        "name",
        "password",
        "email",
        "phone",
        "tenant",
        "queue",
        "status",
    }

    def __init__(
        self,
        name: str,
        password: Optional[str] = configuration.USER_PASSWORD,
        email: Optional[str] = configuration.USER_EMAIL,
        phone: Optional[str] = configuration.USER_PHONE,
        tenant: Optional[str] = configuration.WORKFLOW_TENANT,
        queue: Optional[str] = configuration.WORKFLOW_QUEUE,
        status: Optional[int] = configuration.USER_STATE,
    ):
        super().__init__(name)
        self.user_id: Optional[int] = None
        self.password = password
        self.email = email
        self.phone = phone
        self.tenant = tenant
        self.queue = queue
        self.status = status

    def create_tenant_if_not_exists(self) -> None:
        """Create tenant object."""
        tenant = Tenant(name=self.tenant, queue=self.queue)
        tenant.create_if_not_exists(self.queue)

    def create_if_not_exists(self, **kwargs):
        """Create User if not exists."""
        # Should make sure queue already exists.
        self.create_tenant_if_not_exists()
        user = gateway.create_user(
            self.name,
            self.password,
            self.email,
            self.phone,
            self.tenant,
            self.queue,
            self.status,
        )
        self.user_id = user.getId()
        # TODO recover result checker
        # gateway_result_checker(result, None)

    @classmethod
    def get_user(cls, user_id) -> "User":
        """Get User."""
        user = gateway.query_user(user_id)
        if user is None:
            return cls("")
        user_id = user.getId()
        user = cls(
            name=user.getUserName(),
            password=user.getUserPassword(),
            email=user.getEmail(),
            phone=user.getPhone(),
            tenant=user.getTenantCode(),
            queue=user.getQueueName(),
            status=user.getState(),
        )
        user.user_id = user_id
        return user

    def update(
        self,
        password=None,
        email=None,
        phone=None,
        tenant=None,
        queue=None,
        status=None,
    ) -> None:
        """Update User."""
        user = gateway.update_user(
            self.name,
            password,
            email,
            phone,
            tenant,
            queue,
            status,
        )
        self.user_id = user.getId()
        self.name = user.getUserName()
        self.password = user.getUserPassword()
        self.email = user.getEmail()
        self.phone = user.getPhone()
        self.queue = user.getQueueName()
        self.status = user.getState()

    def delete(self) -> None:
        """Delete User."""
        gateway.delete_user(self.name, self.user_id)
        self.delete_all()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""DolphinScheduler oss resource plugin."""
from typing import Optional
from urllib.parse import urljoin, urlparse

import oss2

from pydolphinscheduler.constants import Symbol
from pydolphinscheduler.core.resource_plugin import ResourcePlugin
from pydolphinscheduler.resources_plugin.base.bucket import Bucket, OSSFileInfo


class OSS(ResourcePlugin, Bucket):
    """OSS object, declare OSS resource plugin for task and workflow to dolphinscheduler.

    :param prefix: A string representing the prefix of OSS.
    :param access_key_id: A string representing the ID of AccessKey for AliCloud OSS.
    :param access_key_secret: A string representing the secret of AccessKey for AliCloud OSS.
    """

    def __init__(
        self,
        prefix: str,
        access_key_id: Optional[str] = None,
        access_key_secret: Optional[str] = None,
        *args,
        **kwargs,
    ):
        super().__init__(prefix, *args, **kwargs)
        self.access_key_id = access_key_id
        self.access_key_secret = access_key_secret

    _bucket_file_info: Optional[OSSFileInfo] = None

    def get_bucket_file_info(self, path: str):
        """Get file information from the file url, like repository name, user, branch, and file path."""
        self.get_index(path, Symbol.SLASH, 3)
        result = urlparse(path)
        hostname = result.hostname
        elements = hostname.split(Symbol.POINT)
        self._bucket_file_info = OSSFileInfo(
            endpoint=f"{result.scheme}://"
            f"{Symbol.POINT.join(str(elements[i]) for i in range(1, len(elements)))}",
            bucket=hostname.split(Symbol.POINT)[0],
            file_path=result.path[1:],
        )

    def read_file(self, suf: str):
        """Get the content of the file.

        The address of the file is the prefix of the resource plugin plus the parameter suf.
        """
        path = urljoin(self.prefix, suf)
        self.get_bucket_file_info(path)
        auth = oss2.Auth(self.access_key_id, self.access_key_secret)
        bucket = oss2.Bucket(
            auth, self._bucket_file_info.endpoint, self._bucket_file_info.bucket
        )
        result = bucket.get_object(self._bucket_file_info.file_path).read().decode()
        return result.read().decode()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Task dvc."""
from copy import deepcopy
from typing import Dict

from pydolphinscheduler.constants import TaskType
from pydolphinscheduler.core.task import Task


class DvcTaskType(str):
    """Constants for dvc task type."""

    INIT = "Init DVC"
    DOWNLOAD = "Download"
    UPLOAD = "Upload"


class BaseDVC(Task):
    """Base class for dvc task."""

    dvc_task_type = None

    _task_custom_attr = {
        "dvc_task_type",
        "dvc_repository",
    }

    _child_task_dvc_attr = set()

    def __init__(self, name: str, repository: str, *args, **kwargs):
        super().__init__(name, TaskType.DVC, *args, **kwargs)
        self.dvc_repository = repository

    @property
    def task_params(self) -> Dict:
        """Return task params."""
        self._task_custom_attr = deepcopy(self._task_custom_attr)
        self._task_custom_attr.update(self._child_task_dvc_attr)
        return super().task_params


class DVCInit(BaseDVC):
    """Task DVC Init object, declare behavior for DVC Init task to dolphinscheduler."""

    dvc_task_type = DvcTaskType.INIT

    _child_task_dvc_attr = {"dvc_store_url"}

    def __init__(self, name: str, repository: str, store_url: str, *args, **kwargs):
        super().__init__(name, repository, *args, **kwargs)
        self.dvc_store_url = store_url


class DVCDownload(BaseDVC):
    """Task DVC Download object, declare behavior for DVC Download task to dolphinscheduler."""

    dvc_task_type = DvcTaskType.DOWNLOAD

    _child_task_dvc_attr = {
        "dvc_load_save_data_path",
        "dvc_data_location",
        "dvc_version",
    }

    def __init__(
        self,
        name: str,
        repository: str,
        data_path_in_dvc_repository: str,
        data_path_in_worker: str,
        version: str,
        *args,
        **kwargs
    ):
        super().__init__(name, repository, *args, **kwargs)
        self.dvc_data_location = data_path_in_dvc_repository
        self.dvc_load_save_data_path = data_path_in_worker
        self.dvc_version = version


class DVCUpload(BaseDVC):
    """Task DVC Upload object, declare behavior for DVC Upload task to dolphinscheduler."""

    dvc_task_type = DvcTaskType.UPLOAD

    _child_task_dvc_attr = {
        "dvc_load_save_data_path",
        "dvc_data_location",
        "dvc_version",
        "dvc_message",
    }

    def __init__(
        self,
        name: str,
        repository: str,
        data_path_in_worker: str,
        data_path_in_dvc_repository: str,
        version: str,
        message: str,
        *args,
        **kwargs
    ):
        super().__init__(name, repository, *args, **kwargs)
        self.dvc_data_location = data_path_in_dvc_repository
        self.dvc_load_save_data_path = data_path_in_worker
        self.dvc_version = version
        self.dvc_message = message



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""DolphinScheduler Project object."""

from typing import Optional

from pydolphinscheduler import configuration
from pydolphinscheduler.java_gateway import gateway
from pydolphinscheduler.models import BaseSide


class Project(BaseSide):
    """DolphinScheduler Project object."""

    def __init__(
        self,
        name: str = configuration.WORKFLOW_PROJECT,
        description: Optional[str] = None,
        code: Optional[int] = None,
    ):
        super().__init__(name, description)
        self.code = code

    def create_if_not_exists(self, user=configuration.USER_NAME) -> None:
        """Create Project if not exists."""
        gateway.create_or_grant_project(user, self.name, self.description)
        # TODO recover result checker
        # gateway_result_checker(result, None)

    @classmethod
    def get_project_by_name(cls, user=configuration.USER_NAME, name=None) -> "Project":
        """Get Project by name."""
        project = gateway.query_project_by_name(user, name)
        if project is None:
            return cls()
        return cls(
            name=project.getName(),
            description=project.getDescription(),
            code=project.getCode(),
        )

    def update(
        self,
        user=configuration.USER_NAME,
        project_code=None,
        project_name=None,
        description=None,
    ) -> None:
        """Update Project."""
        gateway.update_project(user, project_code, project_name, description)
        self.name = project_name
        self.description = description

    def delete(self, user=configuration.USER_NAME) -> None:
        """Delete Project."""
        gateway.delete_project(user, self.code)
        self.delete_all()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# [start workflow_declare]
"""A example workflow for task mlflow."""

from pydolphinscheduler.core.workflow import Workflow
from pydolphinscheduler.tasks.mlflow import (
    MLflowDeployType,
    MLflowModels,
    MLFlowProjectsAutoML,
    MLFlowProjectsBasicAlgorithm,
    MLFlowProjectsCustom,
)

mlflow_tracking_uri = "http://127.0.0.1:5000"

with Workflow(
    name="task_mlflow_example",
    tenant="tenant_exists",
) as workflow:
    # run custom mlflow project to train model
    train_custom = MLFlowProjectsCustom(
        name="train_xgboost_native",
        repository="https://github.com/mlflow/mlflow#examples/xgboost/xgboost_native",
        mlflow_tracking_uri=mlflow_tracking_uri,
        parameters="-P learning_rate=0.2 -P colsample_bytree=0.8 -P subsample=0.9",
        experiment_name="xgboost",
    )

    # run automl to train model
    train_automl = MLFlowProjectsAutoML(
        name="train_automl",
        mlflow_tracking_uri=mlflow_tracking_uri,
        parameters="time_budget=30;estimator_list=['lgbm']",
        experiment_name="automl_iris",
        model_name="iris_A",
        automl_tool="flaml",
        data_path="/data/examples/iris",
    )

    # Using DOCKER to deploy model from train_automl
    deploy_docker = MLflowModels(
        name="deploy_docker",
        model_uri="models:/iris_A/Production",
        mlflow_tracking_uri=mlflow_tracking_uri,
        deploy_mode=MLflowDeployType.DOCKER,
        port=7002,
    )

    train_automl >> deploy_docker

    # run lightgbm to train model
    train_basic_algorithm = MLFlowProjectsBasicAlgorithm(
        name="train_basic_algorithm",
        mlflow_tracking_uri=mlflow_tracking_uri,
        parameters="n_estimators=200;learning_rate=0.2",
        experiment_name="basic_algorithm_iris",
        model_name="iris_B",
        algorithm="lightgbm",
        data_path="/data/examples/iris",
        search_params="max_depth=[5, 10];n_estimators=[100, 200]",
    )

    # Using MLFLOW to deploy model from training lightgbm project
    deploy_mlflow = MLflowModels(
        name="deploy_mlflow",
        model_uri="models:/iris_B/Production",
        mlflow_tracking_uri=mlflow_tracking_uri,
        deploy_mode=MLflowDeployType.MLFLOW,
        port=7001,
    )

    train_basic_algorithm >> deploy_mlflow

    workflow.submit()

# [end workflow_declare]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
This example show how to upload file to dolphinscheduler resource center and use them in tasks.

When you want to create a new resource file in resource center, you can add them to ``workflow.resource_list``
use the following code:

.. code-block:: python

    with Workflow(
        name="use_resource_center",
        resource_list=[
            Resource(name="new-name.py", content="print('hello world from resource center')"),
        ],
        ) as workflow:

during the workflow running, the resource file will be created and uploaded to dolphinscheduler resource
center automatically.

If you want to use the resource file in tasks, you can also use ``resource_list`` parameter in task
constructor, just like the following code:

.. code-block:: python

    task_use_resource = Shell(
        name="run_resource",
        command="python new-name.py",
        resource_list=[
            "new-name.py",
        ],
    )

and the resource file will be downloaded to the task runtime working directory which mean you cna execute
them. In this example we run the file ``new-name.py`` like we execute python script in terminal. And we can
also use the resource file already in dolphinscheduler resource center, not only the new we created in
current workflow.
"""

# [start workflow]
from pydolphinscheduler.core import Workflow
from pydolphinscheduler.core.resource import Resource
from pydolphinscheduler.tasks import Shell

dependence = "dependence.py"
main = "main.py"

with Workflow(
    name="multi_resources_example",
    tenant="tenant_exists",
    # [start create_new_resources]
    resource_list=[
        Resource(
            name=dependence,
            content="from datetime import datetime\nnow = datetime.now()",
        ),
        Resource(name=main, content="from dependence import now\nprint(now)"),
    ],
    # [end create_new_resources]
) as workflow:
    # [start use_exists_resources]
    task_use_resource = Shell(
        name="use-resource",
        command=f"python {main}",
        resource_list=[
            dependence,
            main,
        ],
    )
    # [end use_exists_resources]

    workflow.run()
# [end workflow]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""The script for setting up pydolphinscheduler."""
import logging
import os
import sys
from distutils.command.sdist import sdist
from distutils.dir_util import remove_tree
from distutils.errors import DistutilsExecError
from typing import List

from setuptools import Command, setup

if sys.version_info[0] < 3:
    raise Exception(
        "pydolphinscheduler does not support Python 2. Please upgrade to Python 3."
    )

logger = logging.getLogger(__name__)


class CleanCommand(Command):
    """Command to clean up python api before setup by running `python setup.py clean`."""

    description = "Clean up project root"
    user_options: List[str] = []
    clean_list = [
        "build",
        "htmlcov",
        "dist",
        ".pytest_cache",
        ".coverage",
    ]

    def initialize_options(self) -> None:
        """Set default values for options."""

    def finalize_options(self) -> None:
        """Set final values for options."""

    def run(self) -> None:
        """Run and remove temporary files."""
        for cl in self.clean_list:
            if not os.path.exists(cl):
                logger.info("Path %s do not exists.", cl)
            elif os.path.isdir(cl):
                remove_tree(cl)
            else:
                os.remove(cl)
        logger.info("Finish clean process.")


class ApacheRelease(sdist):
    """Command to make Apache release by running `python setup.py sdist`.

    This command will make a tarball and also sign the tarball with gpg and sha512.
    """

    def run(self):
        """Run and build tarball and sign."""
        super().run()
        version = self.distribution.metadata.get_version()
        target_name = f"dolphinscheduler-python-src-{version}.tar.gz"
        try:
            os.system(
                f"cd dist && "
                f"mv apache-dolphinscheduler-{version}.tar.gz {target_name} && "
                f"gpg --batch --yes --armor --detach-sig {target_name} && "
                f"shasum -a 512 {target_name} > {target_name}.sha512"
            )
        except DistutilsExecError as e:
            self.warn("Make dist and sign failed: %s" % e)


setup(
    cmdclass={
        "clean": CleanCommand,
        "asdist": ApacheRelease,
    },
)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Task SageMaker."""

from pydolphinscheduler.constants import TaskType
from pydolphinscheduler.core.task import Task


class SageMaker(Task):
    """Task SageMaker object, declare behavior for SageMaker task to dolphinscheduler.

    :param name: A unique, meaningful string for the SageMaker task.
    :param sagemaker_request_json: Request parameters of StartPipelineExecution，
        see also `AWS API
        <https://docs.aws.amazon.com/sagemaker/latest/APIReference/API_StartPipelineExecution.html>`_

    """

    _task_custom_attr = {
        "sagemaker_request_json",
    }

    def __init__(self, name: str, sagemaker_request_json: str, *args, **kwargs):
        super().__init__(name, TaskType.SAGEMAKER, *args, **kwargs)
        self.sagemaker_request_json = sagemaker_request_json



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Test pydolphinscheduler docs."""

import re

from tests.testing.constants import task_without_example
from tests.testing.path import get_doc_tasks, get_tasks

# Task sub_process is deprecated and will be removed in the future, so we set it to ignore
ignore_code_file = {"__init__.py", "sub_process.py"}
ignore_doc_file = {"index.rst"}


def test_without_missing_task_rst():
    """Test without missing any task document by compare filename.

    Avoiding add new type of tasks but without adding document about it.
    """
    code_files = {p.stem for p in get_tasks(ignore_name=ignore_code_file)}
    doc_files = {p.stem for p in get_doc_tasks(ignore_name=ignore_doc_file)}
    assert code_files == doc_files


def test_task_without_example():
    """Test task document which without example.

    Avoiding add new type of tasks but without adding example content describe how to use it.
    """
    task_without_example_detected = set()
    pattern = re.compile("Example\n-------")

    for doc in get_doc_tasks(ignore_name=ignore_doc_file):
        search_result = pattern.search(doc.read_text())
        if not search_result:
            task_without_example_detected.add(doc.stem)
    assert task_without_example == task_without_example_detected


def test_doc_automodule_directive_name():
    """Test task document with correct name in directive automodule."""
    pattern = re.compile(".. automodule:: (.*)")
    for doc in get_doc_tasks(ignore_name=ignore_doc_file):
        match_string = pattern.search(doc.read_text()).group(1)
        assert f"pydolphinscheduler.tasks.{doc.stem}" == match_string



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Task Conditions."""

from typing import Dict, List

from pydolphinscheduler.constants import TaskType
from pydolphinscheduler.core.task import Task
from pydolphinscheduler.exceptions import PyDSParamException
from pydolphinscheduler.models.base import Base


class Status(Base):
    """Base class of Condition task status.

    It a parent class for :class:`SUCCESS` and :class:`FAILURE`. Provider status name
    and :func:`get_define` to sub class.
    """

    def __init__(self, *tasks):
        super().__init__(f"Condition.{self.status_name()}")
        self.tasks = tasks

    def __repr__(self) -> str:
        return "depend_item_list"

    @classmethod
    def status_name(cls) -> str:
        """Get name for Status or its sub class."""
        return cls.__name__.upper()

    def get_define(self, camel_attr: bool = True) -> List:
        """Get status definition attribute communicate to Java gateway server."""
        content = []
        for task in self.tasks:
            if not isinstance(task, Task):
                raise PyDSParamException(
                    "%s only accept class Task or sub class Task, but get %s",
                    (self.status_name(), type(task)),
                )
            content.append({"depTaskCode": task.code, "status": self.status_name()})
        return content


class SUCCESS(Status):
    """Class SUCCESS to task condition, sub class of :class:`Status`."""

    def __init__(self, *tasks):
        super().__init__(*tasks)


class FAILURE(Status):
    """Class FAILURE to task condition, sub class of :class:`Status`."""

    def __init__(self, *tasks):
        super().__init__(*tasks)


class ConditionOperator(Base):
    """Set ConditionTask or ConditionOperator with specific operator."""

    _DEFINE_ATTR = {
        "relation",
    }

    def __init__(self, *args):
        super().__init__(self.__class__.__name__)
        self.args = args

    def __repr__(self) -> str:
        return "depend_task_list"

    @classmethod
    def operator_name(cls) -> str:
        """Get operator name in different class."""
        return cls.__name__.upper()

    @property
    def relation(self) -> str:
        """Get operator name in different class, for function :func:`get_define`."""
        return self.operator_name()

    def set_define_attr(self) -> str:
        """Set attribute to function :func:`get_define`.

        It is a wrapper for both `And` and `Or` operator.
        """
        result = []
        attr = None
        for condition in self.args:
            if isinstance(condition, (Status, ConditionOperator)):
                if attr is None:
                    attr = repr(condition)
                elif repr(condition) != attr:
                    raise PyDSParamException(
                        "Condition %s operator parameter only support same type.",
                        self.relation,
                    )
            else:
                raise PyDSParamException(
                    "Condition %s operator parameter support ConditionTask and ConditionOperator but got %s.",
                    (self.relation, type(condition)),
                )
            if attr == "depend_item_list":
                result.extend(condition.get_define())
            else:
                result.append(condition.get_define())
        setattr(self, attr, result)
        return attr

    def get_define(self, camel_attr=True) -> Dict:
        """Overwrite Base.get_define to get task Condition specific get define."""
        attr = self.set_define_attr()
        dependent_define_attr = self._DEFINE_ATTR.union({attr})
        return super().get_define_custom(
            camel_attr=True, custom_attr=dependent_define_attr
        )


class And(ConditionOperator):
    """Operator And for task condition.

    It could accept both :class:`Task` and children of :class:`ConditionOperator`,
    and set AND condition to those args.
    """

    def __init__(self, *args):
        super().__init__(*args)


class Or(ConditionOperator):
    """Operator Or for task condition.

    It could accept both :class:`Task` and children of :class:`ConditionOperator`,
    and set OR condition to those args.
    """

    def __init__(self, *args):
        super().__init__(*args)


class Condition(Task):
    """Task condition object, declare behavior for condition task to dolphinscheduler."""

    def __init__(
        self,
        name: str,
        condition: ConditionOperator,
        success_task: Task,
        failed_task: Task,
        *args,
        **kwargs,
    ):
        super().__init__(name, TaskType.CONDITIONS, *args, **kwargs)
        self.condition = condition
        self.success_task = success_task
        self.failed_task = failed_task
        # Set condition tasks as current task downstream
        self._set_dep()

    def _set_dep(self) -> None:
        """Set upstream according to parameter `condition`."""
        upstream = []
        for cond in self.condition.args:
            if isinstance(cond, ConditionOperator):
                for status in cond.args:
                    upstream.extend(list(status.tasks))
        self.set_upstream(upstream)
        self.set_downstream([self.success_task, self.failed_task])

    @property
    def condition_result(self) -> Dict:
        """Get condition result define for java gateway."""
        return {
            "successNode": [self.success_task.code],
            "failedNode": [self.failed_task.code],
        }

    @property
    def task_params(self, camel_attr: bool = True, custom_attr: set = None) -> Dict:
        """Override Task.task_params for Condition task.

        Condition task have some specials attribute `dependence`, and in most of the task
        this attribute is None and use empty dict `{}` as default value. We do not use class
        attribute `_task_custom_attr` due to avoid attribute cover.
        """
        params = super().task_params
        params["dependence"] = self.condition.get_define()
        return params



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Task function wrapper allows using decorator to create a task."""

import functools
import inspect
import types
from pathlib import Path

from stmdency.extractor import Extractor

from pydolphinscheduler.exceptions import PyDSParamException
from pydolphinscheduler.tasks.python import Python


def _exists_other_decorator(func: types.FunctionType) -> None:
    """Check if the function has other decorators except @task.

    :param func: The function which wraps by decorator ``@task``.
    """
    lines = inspect.getsourcelines(func)[0]

    for line in lines:
        strip_line = line.strip()
        if strip_line.startswith("@") and not strip_line == "@task":
            raise PyDSParamException(
                "Do no support other decorators for function ``task`` decorator."
            )


def task(func: types.FunctionType):
    """Decorate which covert Python functions into pydolphinscheduler task.

    :param func: The function which wraps by decorator ``@task``.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        _exists_other_decorator(func)
        loc = func.__code__.co_filename
        extractor = Extractor(Path(loc).open("r").read())
        stm = extractor.get_code(func.__name__)
        return Python(
            name=kwargs.get("name", func.__name__),
            definition=f"{stm}{func.__name__}()",
            *args,
            **kwargs,
        )

    return wrapper



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""DolphinScheduler ResourcePlugin object."""

from abc import ABCMeta, abstractmethod

from pydolphinscheduler.exceptions import PyResPluginException


# [start resource_plugin_definition]
class ResourcePlugin(object, metaclass=ABCMeta):
    """ResourcePlugin object, declare resource plugin for task and workflow to dolphinscheduler.

    :param prefix: A string representing the prefix of ResourcePlugin.

    """

    # [start init_method]
    def __init__(self, prefix: str, *args, **kwargs):
        self.prefix = prefix

    # [end init_method]

    # [start abstractmethod read_file]
    @abstractmethod
    def read_file(self, suf: str):
        """Get the content of the file.

        The address of the file is the prefix of the resource plugin plus the parameter suf.
        """

    # [end abstractmethod read_file]

    def get_index(self, s: str, x, n):
        """Find the subscript of the nth occurrence of the X character in the string s."""
        if n <= s.count(x):
            all_index = [key for key, value in enumerate(s) if value == x]
            return all_index[n - 1]
        else:
            raise PyResPluginException("Incomplete path.")


# [end resource_plugin_definition]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# [start workflow_declare]
r"""
A example workflow for task condition.

This example will create five task in single workflow, with four shell task and one condition task. Task
condition have one upstream which we declare explicit with syntax `parent >> condition`, and three downstream
automatically set dependence by condition task by passing parameter `condition`. The graph of this workflow
like:
pre_task_1 ->                     -> success_branch
             \                  /
pre_task_2 ->  -> conditions ->
             /                  \
pre_task_3 ->                     -> fail_branch
.
"""

from pydolphinscheduler.core.workflow import Workflow
from pydolphinscheduler.tasks.condition import FAILURE, SUCCESS, And, Condition
from pydolphinscheduler.tasks.shell import Shell

with Workflow(name="task_condition_example", tenant="tenant_exists") as workflow:
    pre_task_1 = Shell(name="pre_task_1", command="echo pre_task_1")
    pre_task_2 = Shell(name="pre_task_2", command="echo pre_task_2")
    pre_task_3 = Shell(name="pre_task_3", command="echo pre_task_3")
    cond_operator = And(
        And(
            SUCCESS(pre_task_1, pre_task_2),
            FAILURE(pre_task_3),
        ),
    )

    success_branch = Shell(name="success_branch", command="echo success_branch")
    fail_branch = Shell(name="fail_branch", command="echo fail_branch")

    condition = Condition(
        name="condition",
        condition=cond_operator,
        success_task=success_branch,
        failed_task=fail_branch,
    )
    workflow.submit()
# [end workflow_declare]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""String util function collections."""

from pydolphinscheduler.constants import Delimiter


def attr2camel(attr: str, include_private=True):
    """Covert class attribute name to camel case."""
    if include_private:
        attr = attr.lstrip(Delimiter.UNDERSCORE)
    return snake2camel(attr)


def snake2camel(snake: str):
    """Covert snake case to camel case."""
    components = snake.split(Delimiter.UNDERSCORE)
    return components[0] + "".join(x.title() for x in components[1:])


def class_name2camel(class_name: str):
    """Covert class name string to camel case."""
    class_name = class_name.lstrip(Delimiter.UNDERSCORE)
    return class_name[0].lower() + snake2camel(class_name[1:])
